class Postprocessor:
    def __init__(self, base_tagger, stopwords=None, passwords=None, only_tags=None, replace=None):
        self.base_tagger = base_tagger
        self.stopwords = stopwords if stopwords else None
        self.passwords = passwords if passwords else None
        self.only_tags = only_tags if only_tags else None
        self.replace = replace if replace else None
        
    def tag(self, phrase):
        def to_replace(w):
            if w in self.replace: return self.replace[w]
            if w[0] in self.replace: return (self.replace[w[0]], w[1])
            return w

        words = self.base_tagger.pos(phrase)
        if self.stopwords:
            words = [w for w in words if not ((w in self.stopwords) or (w[0] in self.stopwords))]
        if self.passwords:
            words = [w for w in words if ((w in self.passwords) or (w[0] in self.passwords))]
        if self.only_tags:
            words = [w for w in words if w[1] in self.only_tags]
        if self.replace:
            words = [to_replace(w) for w in words]
        return words


import glob
import os


installpath = os.path.dirname(os.path.realpath(__file__))

def load_dictionary(directory, encoding='utf-8', ignore_a_syllable=False):
    fnames = glob.glob('%s/*.txt' % directory)
    words = set()
    for fname in fnames:
        try:
            with open(fname, encoding=encoding) as f:
                words_ = {line.strip() for line in f}
                if ignore_a_syllable:
                    words_ = {w for w in words_ if len(w) > 1}
                words.update(words_)
        except Exception as e:
            print(e)
            continue
    return words

def loadtxt(fname, encoding='utf-8'):
    try:
        with open(fname, encoding=encoding) as f:
            return [line.strip() for line in f]
    except Exception as e:
        print(e)
        return []


from konlpy.tag import Twitter as KoNLPyTwitter
from ckonlpy.custom_tag import SimpleTemplateTagger
from ckonlpy.custom_tag import SimpleSelector
from ckonlpy.data.tagset import twitter as tagset
from ckonlpy.dictionary import CustomizedDictionary
from ckonlpy.utils import installpath
from ckonlpy.utils import load_dictionary
from ckonlpy.utils import loadtxt


class Twitter:
    def __init__(self, load_default_dictionary=True):
        self._base = KoNLPyTwitter()
        self._dictionary = CustomizedDictionary()
        if load_default_dictionary:
            self._load_default_dictionary()
        self._customized_tagger = self._load_customized_tagger()
        self.tagset = tagset
    
    def _load_customized_tagger(self):        
        templatespath = '%s/data/templates/twitter_templates0' % installpath
        templates = loadtxt(templatespath)
        templates = [tuple(template.split()) for template in templates]        
        return SimpleTemplateTagger(templates, self._dictionary, SimpleSelector())
    
    def _load_default_dictionary(self):
        directory = '%s/data/twitter/' % installpath
        self._dictionary.add_dictionary(load_dictionary('%s/josa' % directory), 'Josa')
        self._dictionary.add_dictionary(load_dictionary('%s/noun' % directory, ignore_a_syllable=True), 'Noun')
        self._dictionary.add_dictionary(load_dictionary('%s/adverb' % directory), 'Adverb')
        #self._dictionary.add_dictionary(load_dictionary(modifier_dir), 'Modifier')
        
    def pos(self, phrase):
        eojeols = phrase.split()
        tagged = []
        for eojeol in eojeols:
            tagged0 = self._customized_tagger.pos(eojeol)
            if tagged0:
                tagged += tagged0
                continue
            tagged += self._base.pos(eojeol)
        return tagged
    
    def nouns(self, phrase):
        tagged = self.pos(phrase)
        return [w for w, t in tagged if t[0] == 'N']
    
    def morphs(self, phrase, norm=False, stem=False):
        return [s for s, t in self.pos(phrase)]
    
    def phrases(self, phrase):
        # TODO
        return self._base.phrases(phrase)
    
    def add_dictionary(self, words, tag, force=False):
        if (not force) and (not (tag in self.tagset)):
            raise ValueError('%s is not available tag' % tag)
        self._dictionary.add_dictionary(words, tag)
    
    def load_dictionary(self, fname_list, tag):
        if not (tag in self.tagset):
            raise ValueError('%s is not available tag' % tag)
        self._dictionary.load_dictionary(fname_list, tag)
    
    def set_selector(self, my_weight_dict, my_score_function):
        self._customized_tagger.set_selector(my_weight_dict, my_score_function)


class SimpleTemplateTagger:
    def __init__(self, templates, dictionary, selector=None, max_length=10):
        self.templates = templates
        self.dictionary = dictionary
        self.max_length = max_length
        self.selector = selector if selector else SimpleSelector()

    def pos(self, eojeol):
        """eojeol: str"""

        for template in self.templates:
            n = len(eojeol)
            
            # Initialize
            candidates = []
            for e in range(1, min(self.max_length, n)+1):
                word = eojeol[:e]
                if self.dictionary.is_tag(word, template[0]):
                    candidates.append([(word, template[0], e)])
            if not candidates:
                continue

            # Expansion
            for t in template[1:]:
                candidates_ = []
                for candidate in candidates:
                    word, t0, b = candidate[-1]
                    for e in range(b, min(self.max_length+b, n)+1):
                        word = eojeol[b:e]
                        if self.dictionary.is_tag(word, t):
                            candidates_.append(candidate + [(word, t, e)])
                candidates = candidates_
                if not candidates:
                    break
            
            # Select best one
            candidates = [[tagged[:2] for tagged in c] for c in candidates if c[-1][2] == n]
            if candidates:
                return self.selector.select(candidates)

        return None
    
    def add_a_template(self, a_template):
        if type(a_template) != tuple:
            a_template = tuple(a_template)
        if (a_template in self.templates) == False:
            self.templates.append(a_template)
    
    def set_selector(self, my_weight_dict, my_score_function):
        self.selector.weight = my_weight_dict
        self.selector.score = my_score_function
        test_candidates = [
            [('이', 'Noun'), ('것', 'Noun'), ('은', 'Josa'), ('테', 'Noun'), ('스트', 'Noun')],
            [('이것', 'Noun'), ('은', 'Josa'), ('테', 'Noun'), ('스트', 'Noun')],
            [('이것', 'Noun'), ('은', 'Josa'), ('테스트', 'Noun')]
        ]
        for test_candidate in test_candidates:
            print(test_candidate)
        print('best:', self.selector.select(test_candidates))

class SimpleSelector:
    def __init__(self):
        self.weight = {
            'noun_length': 1,
            'noun_numbers': -0.2,
            'wordlist_length': -0.1,
            'no_noun': -0.5
        }

    def select(self, candidates):
        scores = []
        for candidate in candidates:
            score = self.score(candidate)
            scores.append(score)
        best = sorted(enumerate(scores), key=lambda x:x[1], reverse=True)[0][0]
        return candidates[best]
    
    def score(self, candidate):
        noun_length = sum([len(w) for w, t in candidate if t == 'Noun'])
        noun_numbers = len([1 for w, t in candidate if t == 'Noun'])
        wordlist_length = len(candidate)
        no_noun = 0 if [1 for w, t in candidate if t == 'Noun'] else 1

        return noun_length * self.weight.get('noun_length', 0) + noun_numbers * self.weight.get('noun_numbers', 0) + wordlist_length * self.weight.get('wordlist_length', 0) + no_noun * self.weight.get('no_noun', 0)


from description import __version__, __author__
from setuptools import setup, find_packages

setup(
   name="customized_KoNLPy",
   version=__version__,
   author=__author__,
   author_email='soy.lovit@gmail.com',
   url='https://github.com/lovit/customized_konlpy',
   description="KoNLPy customization version (wrapping package)",
   long_description="""Add customized words and tags and pos tagging with templates   
   """,
   install_requires=["Jpype1>=0.6.1", "konlpy>=0.4.4"],
   keywords = ['KoNLPy wrapping customization'],
   packages=find_packages(),
   package_data={'ckonlpy':['data/*/*.txt', 'data/*/*/*.txt']}
)